# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class TronFrozenSupply(p.MessageType):

    def __init__(
        self,
        frozen_amount: int = None,
        frozen_days: int = None,
    ) -> None:
        self.frozen_amount = frozen_amount
        self.frozen_days = frozen_days

    @classmethod
    def get_fields(cls):
        return {
            1: ('frozen_amount', p.UVarintType, 0),
            2: ('frozen_days', p.UVarintType, 0),
        }
