# Automatically generated by pb2py
# fmt: off
from .. import protobuf as p


class TronTransferAssetContract(p.MessageType):

    def __init__(
        self,
        asset_name: str = None,
        to_address: bytes = None,
        amount: int = None,
    ) -> None:
        self.asset_name = asset_name
        self.to_address = to_address
        self.amount = amount

    @classmethod
    def get_fields(cls):
        return {
            1: ('asset_name', p.UnicodeType, 0),
            2: ('to_address', p.BytesType, 0),
            3: ('amount', p.UVarintType, 0),
        }
